/*
 *   Copyright (C) 2003-2004 Fredrik Lingvall (fl@signal.uu.se)
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "config.h"

#include "i18n.h"

#include <xmms/plugin.h>
#include <xmms/util.h>

#include <glib.h>
#include <stdlib.h>
#include <pthread.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <string.h>
#include <stdio.h>
#include <glob.h>
#include <math.h>

#define MIN_FREQ 10
#define MAX_FREQ 20000


// Linux sound driver constants.
#define CHANNELS  2
#define RATE      44100
#define AFMT      AFMT_S16_LE
#define FRAG      0x7fff000b // unlimited number of fragments of size 2^11 = 256 bytes.
//#define FRAG      0x0004000a;
//#define FRAG_SIZE 2048
#define FRAG_SIZE 512



// Need to check this up...

/*
  sscanf(Tcl_line + 12,"%hd",&x);
  close(Audio_fd);
  Audio_arg = 0x00040000 + x;
  Audio_BufSize = 1<<x-1;	
  opencard();
*/


// XMMS sound constants.
#define XMMS_FMT FMT_S16_LE


//#define FRAG_SIZE 4096
#define BUF_BYTES FRAG_SIZE

#define BUF_SAMPLES BUF_BYTES/2

#define AUDIO_ERR 1
#define AUDIO_OK  0

#define  STRLEN 255


typedef struct
{
  int count;   // Number of available devices.
  char **name; // List of device names.
} AudioDevices;


static InputPlugin sc_ip;

static gboolean  going;
static gboolean  audio_error;
static pthread_t play_thread;
static gint      fd_audio;	// Audio device file descriptor.
static gint      channels, rate, format; // Audio parameters.


/********************************* Soundcard stuff ************************************************/

/***
 *
 * Find a list of available audio devices.
 *
 ***/

static void scan_audio_devices(AudioDevices *devices, gchar* infostr)
{
  gint    i,j;
  gint    mask, fd;
  glob_t result;
  //AudioDevices *devices;

  gint channelsT = CHANNELS;
  gint rateT     = RATE;
  gint afmtT     = AFMT;
  gint frag      = FRAG;
  
  if (glob("/dev/dsp*", 0, NULL, &result) != 0) {
    devices = NULL;
    return;
  }
  
  if (result.gl_pathc < 1) {
    devices = NULL;
    return;
  }
  
  sprintf(infostr,"%d dsp devices found.\n",result.gl_pathc);
  sprintf(infostr,"Required: Channels = %d, speed = %d, frag = %d\n",channels,rate,FRAG);

  for (i = 0, j = 0; i < result.gl_pathc; i++) {
    
    fd = open(result.gl_pathv[i], O_RDONLY, 0);
    
    ioctl(fd, SNDCTL_DSP_GETFMTS, &mask);
    ioctl(fd, SNDCTL_DSP_CHANNELS, &channelsT);
    ioctl(fd, SNDCTL_DSP_SPEED, &rateT);
    ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &frag);
    sprintf(infostr,"device = %s, fd = %d [%d], FMTS= %d, Channels = %d, speed = %d, frag = %d\n",
	    result.gl_pathv[i],fd,i,mask,channelsT,rateT,frag);
    
    if (fd != -1)
      close (fd);
  }
  
  devices       = malloc(sizeof(AudioDevices));
  devices->name = malloc( (sizeof (gchar*)) * result.gl_pathc);
  
  for (i = 0, j = 0; i < result.gl_pathc; i++)
    {
    
      // Scan audio device.

      if ((fd = open(result.gl_pathv[i], O_RDONLY, 0)) != -1 &&
	  ioctl(fd, SNDCTL_DSP_GETFMTS, &mask)         != -1 &&
	  (mask & AFMT)                                      &&
	  ioctl(fd, SNDCTL_DSP_SETFMT, &afmtT)         != -1 &&
	  afmtT == AFMT                                      &&
	  ioctl(fd, SNDCTL_DSP_CHANNELS, &channelsT)   != -1 &&
	  channelsT == CHANNELS                              &&
	  ioctl(fd, SNDCTL_DSP_SPEED, &rateT)          != -1 &&
	  rateT == RATE                                      &&
	  ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &frag)     != -1)
	{
	  devices->name[j] = malloc(strlen(result.gl_pathv[i])-5+1);
	  strcpy(devices->name[j], result.gl_pathv[i]+5);
	  sprintf(infostr,"Audio devices found : %s\n",devices->name[j]);
	  j++;
	}
      close(fd);
    }

  devices->count = j;

  return;
}

/***
 *
 * Open the audio device and set the sampling rate etc.
 *
 ***/

static gint open_sounddevice(gchar *device)
{
  gint  mask;
  gint  channelsT = channels;
  gint  rateT     = rate;
  gint  afmtT     = format;
  gint  frag      = FRAG;
  gint  caps;
  gchar thestr[STRLEN]; 


  printf("open_sounddevice : channels = %d, rate = %d, format = %d\n",channels,rate,format);

  if (device == NULL)
    return AUDIO_ERR;
  
  // Open as read-only.
  if ((fd_audio = open(device, O_RDONLY, 0)) == -1) {
    perror("Open audio device failed");   
    going = FALSE;
    return(AUDIO_ERR);
  }


  //
  // Check and set audio format.
  //
  if (ioctl(fd_audio, SNDCTL_DSP_GETFMTS, &mask) == -1) {
    perror("Getting format info from soundcard failed");
    going = FALSE;
    return(AUDIO_ERR);
  }
  
  if (!(mask & AFMT)) {
    fprintf(stderr,"Format not supported by soundcard");
    going = FALSE;
    return(AUDIO_ERR);
  }

  if (ioctl(fd_audio, SNDCTL_DSP_SETFMT, &afmtT) == -1) {
    perror("Set audio format failed");
    going = FALSE;
    return(AUDIO_ERR);
  }

  if (afmtT != AFMT) {
    fprintf(stderr,"Format not supported by soundcard");
    going = FALSE;
    return(AUDIO_ERR);
  }

  //
  // Set number of channels.
  //
  
  if (ioctl(fd_audio, SNDCTL_DSP_CHANNELS, &channelsT) == -1 || channelsT != channels) {
    perror("Setting  no of channels failed)");
    going = FALSE;
    return(AUDIO_ERR);  
  }
  
  //
  // Set sampling rate.
  //

  //if (ioctl(fd_audio, SNDCTL_DSP_SPEED, &rateT)  == -1 || rateT != rate)
  if (ioctl(fd_audio, SNDCTL_DSP_SPEED, &rateT)  == -1 || rateT != rate) { 
    sprintf(thestr,"Setting sampling rate failed %d",rateT);
    perror(thestr);
    going = FALSE;
    return(AUDIO_ERR);
  }
  
  //
  // Set fragment size.
  //

  if (ioctl(fd_audio, SNDCTL_DSP_SETFRAGMENT, &frag)  == -1) {
    perror("Setting fragment size failed");
    going = FALSE;
    return(AUDIO_ERR);
  }

  //
  // Check if soundcard can do full duplex.
  //
#if 0  
  if (ioctl(fd_audio, SNDCTL_DSP_SETDUPLEX, 0)==-1) {
    perror("Setting duplex on audio");
    //close(fd);
    going = FALSE;
    return(AUDIO_ERR);
  }
  
  if (ioctl(fd_audio, SNDCTL_DSP_GETCAPS, &caps)==-1) {
    perror("Can't check card's capabilities");
    //perror("Getting capabilities of audio card failed");
    //close(fd_audio);
    going = FALSE;
    return(AUDIO_ERR);
  }
  
  if (!(caps & DSP_CAP_DUPLEX)) {
    //fprintf(stderr, "Sorry but your soundcard can't do full duplex!\n");
    fprintf(stderr,"Full duplex not supported by soundcard\n");
    //close(fd_audio);
    going = FALSE;
    return(AUDIO_ERR);
  }
#endif
  printf("open_sounddevice : Sound device open.\n");
  
  return(AUDIO_OK);
}

/***
 *
 * Close the audio device.
 *
 ***/

static gint close_sounddevice()
{
  gint err;
  
  err = close(fd_audio);
  if (err) {
    perror("Closing audio device failed");
    return(AUDIO_ERR);
  }
  else
    return(AUDIO_OK);
}

/********************************* Plugin stuff ************************************************/

/***
 *
 * About dialog.
 *
 ***/

static void sc_about(void)
{
  static GtkWidget *box;
  box = xmms_show_message(
			  _("About Soundcard Plugin"),
			  _("Soundcard input plugin by Fredrik Lingvall <fl@signal.uu.se>\n"
			    "Version " VERSION
			    "\nTo use it, add a URL: soundcard://device\n"
			    "e.g. soundcard:///dev/dsp to play to play from the /dev/dsp device"),
			  _("Ok"), FALSE, NULL, NULL);

  gtk_signal_connect(GTK_OBJECT(box), "destroy",
		     gtk_widget_destroyed, &box);
}

/***
 *
 * About dialog.
 *
 ***/

static int sc_is_our_file(char *filename)
{
  if (!strncmp(filename, "soundcard://", 12))
    return TRUE;
  return FALSE;
}

/***
 *
 * 
 *
 ***/	

static void* play_loop(void *arg)
{
  //GArray               *frequencies = arg;
  gchar                *device = arg;
  gint16               data[BUF_BYTES];
  gint                 i;

  unsigned char         buffer_raw[FRAG_SIZE], tmp_buffer[FRAG_SIZE];
  gint                  count,retval;
  fd_set                reads;
  struct audio_buf_info info;
  //gchar                 device[STRLEN];
  //AudioDevices          *ad; 
  struct timeval        time;
  div_t                 d;
  
  printf("%s\n", device);


  for (i=0; i<BUF_BYTES; i++)
    data[i] = 0;

  
  // Open audio device.
  if (open_sounddevice(device) == AUDIO_ERR) {
    going = FALSE;
  }
   
  if (going) {
    // Dummy read - otherwise select() won't work. 
    if (read(fd_audio, tmp_buffer, FRAG_SIZE) != FRAG_SIZE) {
      perror("Read error from sound device");
      going = FALSE;
      close_sounddevice();
    }
    
    FD_ZERO(&reads); 
    count = 0;
    printf("Recording...\n");
  }

  while (going) {
    
    while (count <  BUF_SAMPLES) {
      
      time.tv_sec = 1; // Wait max one second for audio input.
      time.tv_usec= 0;
      
      FD_SET(fd_audio, &reads);
      retval = select(fd_audio+1, &reads, NULL, NULL, &time); // +1 see man page on select.
      if (retval==0) {
	perror("Select error from sound device");
	going = FALSE;
	close_sounddevice();
      }
      
      if (FD_ISSET(fd_audio, &reads)) {
	
	#if 0
	// Print whats happening...
	printf("Audio data is coming in from dev %d\n",fd_audio);
	info.bytes = 0;
	while (info.bytes < FRAG_SIZE) {
	  if (ioctl(fd_audio, SNDCTL_DSP_GETISPACE, &info)==-1) {
	    perror("Audio device read (ioctl) error!\n");
	    close_sounddevice();
	  }
	  printf("Got %d bytes\n",info.bytes);
	}
	#endif

	// Read data from the device.
	
	if (read(fd_audio, buffer_raw, FRAG_SIZE) != FRAG_SIZE) {
	  perror("Audio device read error!\n");
	  going = FALSE;
	  close_sounddevice();
	}
	
	// Write to output buffer.
	for (i = 0; i < FRAG_SIZE-1; i+=2) {

	  data[count] = (gint16) ( buffer_raw[i] | (buffer_raw[i+1]<<8) );
	  count++;
	}
	
	FD_CLR(fd_audio, &reads); // Wait for more audio data.
      }

    } // while (count ...

    // Reset counter. 
    count = 0; 

    // Add data to PCM stream. // NOTE  FMT_S16_LE !=  AFMT_S16_LE
    sc_ip.add_vis_pcm(sc_ip.output->written_time(), XMMS_FMT, channels, BUF_BYTES, data);

    // Wait until ...
    while (sc_ip.output->buffer_free() < BUF_BYTES && going) {
      xmms_usleep(10000);
    }
    
    if (going)
      sc_ip.output->write_audio(data, BUF_BYTES);
    
  } // while (going)
  
  printf("done.\n");

  // Free memory allocated in sc_filename_parse.
  g_free(device);
  //g_array_free(frequencies, TRUE);
  
  // Make sure the output plugin stops prebuffering.
  sc_ip.output->buffer_free();
  sc_ip.output->buffer_free();

  // Close the audio device.
  printf("Closing %s...",device);
  close_sounddevice();  
  
  pthread_exit(NULL);
}

/***
 *
 * 
 *
 ***/	

//static GArray* sc_filename_parse(const char* filename)
static char* sc_filename_parse(const char* filename)
{
  //GArray *device = g_array_new(FALSE, FALSE, sizeof(double));
  gchar *device = g_malloc0(STRLEN);
  //char **strings, **ptr;

  if (strncmp(filename,"soundcard://", 12))
    return NULL;

  filename += 12;
  
  //strings = g_strsplit(filename, ";", 100);

  //for (ptr = strings; *ptr != NULL; ptr++)
  //  {
  //    double freq = strtod(*ptr, NULL);
  //    if (freq >= MIN_FREQ && freq <= MAX_FREQ)
  //	g_array_append_val(device, freq);
  //  }
  //g_strfreev(strings);

  //if (device->len == 0)
  //  {
  //    g_array_free(device, TRUE);
  //    device = NULL;
  //  }

  strcpy(device,filename);

  return device;
}

/***
 *
 * 
 *
 ***/	

static char* sc_title(char *filename)
{
  //GArray *device;
  gchar *device;
  char* title;
  //int i;

  device = sc_filename_parse(filename);
  if (device == NULL)
    return NULL;

  printf("sc_title : %s\n",device);

  //title = g_strdup_printf("%s %.1f Hz", _("Soundcard: "),
  //			  g_array_index(device, double, 0));

  title = g_strdup_printf("%s %s", _("Soundcard: "),device);

  printf("sc_title : %s\n",title);

  //for (i = 1; i < device->len; i++) {
  //  char *old_title;
  //  old_title = title;
  //  title = g_strdup_printf("%s;%.1f Hz", old_title,
  //			    g_array_index(device, double, i));
  //  g_free(old_title);
  //}

  // Free memory allocated in sc_filename_parse.
  g_free(device);

  //g_array_free(device, TRUE);
  
  return title;
}


/***
 *
 * 
 *
 ***/	

static void sc_play(char *filename)
{
  //GArray* device;
  gchar* device;
  char *name;

  // Stereo, 16 bit audio.
  channels = 2;
  format   = AFMT;
  rate     = RATE;   

  device = sc_filename_parse(filename);
  if (device == NULL)
    return;

  printf("sc_play : device = %s\n",device);

  printf("sc_play : format = %d, rate = %d, channels = %d\n",format, rate, channels);

  going = TRUE;
  audio_error = FALSE;
  if (sc_ip.output->open_audio(XMMS_FMT, rate, channels) == 0) {    
    audio_error = TRUE;
    going = FALSE;
    return;
  }

  name = sc_title(filename);

  printf("sc_play : name = %s\n",name);

  sc_ip.set_info(name, -1, 16 * rate, rate, channels);
  g_free(name);
  
  printf("Starting play thread...\n");

  //Start the play thread.
  pthread_create(&play_thread, NULL, play_loop, device);
}

/***
 *
 * 
 *
 ***/	

static void sc_stop(void)
{
  if (going) {
    going = FALSE;
    pthread_join(play_thread, NULL);
    sc_ip.output->close_audio();
  }
}

/***
 *
 * 
 *
 ***/	

static void sc_pause(short paused)
{
  sc_ip.output->pause(paused);
}

/***
 *
 * 
 *
 ***/	

static int sc_get_time(void)
{
  if (audio_error)
    return -2;

  if (!going && !sc_ip.output->buffer_playing())
    return -1;

  return sc_ip.output->output_time();
}

/***
 *
 * 
 *
 ***/	

static void sc_song_info(char *filename, char **title, int *length)
{
  *length = -1;
  *title = sc_title(filename);
}

/***
 *
 * 
 *
 ***/	

static InputPlugin sc_ip = 
  {
    NULL,
    NULL,
    NULL, /* Description */
    NULL,
    sc_about,
    NULL,
    sc_is_our_file,
    NULL,
    sc_play,
    sc_stop,
    sc_pause,
    NULL,
    NULL,
    sc_get_time,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    sc_song_info,
    NULL,
    NULL
  };

/***
 *
 * 
 *
 ***/	

InputPlugin *get_iplugin_info(void)
{
  sc_ip.description = g_strdup_printf(_("Soundcard %s"), VERSION);
  return &sc_ip;
}
